@@ -54,9 +54,12 @@
 #include <private/qtextimagehandler_p.h>
 #include <private/qrawfont_p.h>
 #include <private/qglyphrun_p.h>
+#include <private/qquickitem_p.h>
 
 QT_BEGIN_NAMESPACE
 
+Q_DECLARE_LOGGING_CATEGORY(lcSgText)
+
 QQuickTextNodeEngine::BinaryTreeNodeKey::BinaryTreeNodeKey(BinaryTreeNode *node)
     : fontEngine(QRawFontPrivate::get(node->glyphRun.rawFont())->fontEngine)
     , clipNode(node->clipNode)
@@ -70,7 +73,7 @@ QQuickTextNodeEngine::BinaryTreeNode::BinaryTreeNode(const QGlyphRun &g,
                                                      const QRectF &brect,
                                                      const Decorations &decs,
                                                      const QColor &c,
-                                                     const QColor &bc,
+                                                     const QColor &bc, const QColor &dc,
                                                      const QPointF &pos, qreal a)
     : glyphRun(g)
     , boundingRect(brect)
@@ -79,6 +82,7 @@ QQuickTextNodeEngine::BinaryTreeNode::BinaryTreeNode(const QGlyphRun &g,
     , decorations(decs)
     , color(c)
     , backgroundColor(bc)
+    , decorationColor(dc)
     , position(pos)
     , ascent(a)
     , leftChildIndex(-1)
@@ -91,7 +95,7 @@ QQuickTextNodeEngine::BinaryTreeNode::BinaryTreeNode(const QGlyphRun &g,
 
 void QQuickTextNodeEngine::BinaryTreeNode::insert(QVarLengthArray<BinaryTreeNode, 16> *binaryTree, const QGlyphRun &glyphRun, SelectionState selectionState,
                                              Decorations decorations, const QColor &textColor,
-                                             const QColor &backgroundColor, const QPointF &position)
+                                             const QColor &backgroundColor, const QColor &decorationColor, const QPointF &position)
 {
     QRectF searchRect = glyphRun.boundingRect();
     searchRect.translate(position);
@@ -111,6 +115,7 @@ void QQuickTextNodeEngine::BinaryTreeNode::insert(QVarLengthArray<BinaryTreeNode
                                       decorations,
                                       textColor,
                                       backgroundColor,
+                                      decorationColor,
                                       position,
                                       ascent));
 }
@@ -251,6 +256,7 @@ void QQuickTextNodeEngine::processCurrentLine()
 
     QColor lastColor;
     QColor lastBackgroundColor;
+    QColor lastDecorationColor;
 
     QVarLengthArray<TextDecoration> pendingUnderlines;
     QVarLengthArray<TextDecoration> pendingOverlines;
@@ -265,11 +271,10 @@ void QQuickTextNodeEngine::processCurrentLine()
                 Q_ASSERT(sortedIndex < m_currentLineTree.size());
 
                 node = m_currentLineTree.data() + sortedIndex;
+                if (i == 0)
+                    currentSelectionState = node->selectionState;
             }
 
-            if (i == 0)
-                currentSelectionState = node->selectionState;
-
             // Update decorations
             if (currentDecorations != Decoration::NoDecoration) {
                 decorationRect.setY(m_position.y() + m_currentLine.y());
@@ -279,6 +284,12 @@ void QQuickTextNodeEngine::processCurrentLine()
                     decorationRect.setRight(node->boundingRect.left());
 
                 TextDecoration textDecoration(currentSelectionState, decorationRect, lastColor);
+                if (lastDecorationColor.isValid() &&
+                        (currentDecorations.testFlag(Decoration::Underline) ||
+                         currentDecorations.testFlag(Decoration::Overline) ||
+                         currentDecorations.testFlag(Decoration::StrikeOut)))
+                    textDecoration.color = lastDecorationColor;
+
                 if (currentDecorations & Decoration::Underline)
                     pendingUnderlines.append(textDecoration);
 
@@ -397,6 +408,7 @@ void QQuickTextNodeEngine::processCurrentLine()
                 currentDecorations = node->decorations;
                 lastColor = node->color;
                 lastBackgroundColor = node->backgroundColor;
+                lastDecorationColor = node->decorationColor;
                 m_processedNodes.append(*node);
             }
         }
@@ -476,6 +488,9 @@ void QQuickTextNodeEngine::addTextObject(const QTextBlock &block, const QPointF
             }
         }
 
+        // Use https://developer.mozilla.org/de/docs/Web/CSS/vertical-align as a reference
+        // The top/bottom positions are supposed to be higher/lower than the text and reference
+        // the line height, not the text height (using QFontMetrics)
         qreal ascent;
         QTextLine line = block.layout()->lineForTextPosition(pos - block.position());
         switch (format.verticalAlignment())
@@ -483,11 +498,10 @@ void QQuickTextNodeEngine::addTextObject(const QTextBlock &block, const QPointF
         case QTextCharFormat::AlignTop:
             ascent = line.ascent();
             break;
-        case QTextCharFormat::AlignMiddle: {
-            QFontMetrics m(format.font());
-            ascent = (size.height() - m.xHeight()) / 2;
+        case QTextCharFormat::AlignMiddle:
+            //       Middlepoint of line (height - descent) + Half object height
+            ascent = (line.ascent() + line.descent()) / 2 - line.descent() + size.height() / 2;
             break;
-        }
         case QTextCharFormat::AlignBottom:
             ascent = size.height() - line.descent();
             break;
@@ -508,6 +522,7 @@ void QQuickTextNodeEngine::addUnselectedGlyphs(const QGlyphRun &glyphRun)
                            Decoration::NoDecoration,
                            m_textColor,
                            m_backgroundColor,
+                           m_decorationColor,
                            m_position);
 }
 
@@ -520,6 +535,7 @@ void QQuickTextNodeEngine::addSelectedGlyphs(const QGlyphRun &glyphRun)
                            Decoration::NoDecoration,
                            m_textColor,
                            m_backgroundColor,
+                           m_decorationColor,
                            m_position);
     m_hasSelection = m_hasSelection || m_currentLineTree.size() > currentSize;
 }
@@ -532,12 +548,12 @@ void QQuickTextNodeEngine::addGlyphsForRanges(const QVarLengthArray<QTextLayout:
     int remainingLength = end - start;
     for (int j=0; j<ranges.size(); ++j) {
         const QTextLayout::FormatRange &range = ranges.at(j);
-        if (range.start + range.length >= currentPosition
+        if (range.start + range.length > currentPosition
                 && range.start < currentPosition + remainingLength) {
 
             if (range.start > currentPosition) {
                 addGlyphsInRange(currentPosition, range.start - currentPosition,
-                                 QColor(), QColor(), selectionStart, selectionEnd);
+                                 QColor(), QColor(), QColor(), selectionStart, selectionEnd);
             }
             int rangeEnd = qMin(range.start + range.length, currentPosition + remainingLength);
             QColor rangeColor;
@@ -549,8 +565,12 @@ void QQuickTextNodeEngine::addGlyphsForRanges(const QVarLengthArray<QTextLayout:
                     ? range.format.background().color()
                     : QColor();
 
+            QColor rangeDecorationColor = range.format.hasProperty(QTextFormat::TextUnderlineColor)
+                    ? range.format.underlineColor()
+                    : QColor();
+
             addGlyphsInRange(range.start, rangeEnd - range.start,
-                             rangeColor, rangeBackgroundColor,
+                             rangeColor, rangeBackgroundColor, rangeDecorationColor,
                              selectionStart, selectionEnd);
 
             currentPosition = range.start + range.length;
@@ -562,14 +582,14 @@ void QQuickTextNodeEngine::addGlyphsForRanges(const QVarLengthArray<QTextLayout:
     }
 
     if (remainingLength > 0) {
-        addGlyphsInRange(currentPosition, remainingLength, QColor(), QColor(),
+        addGlyphsInRange(currentPosition, remainingLength, QColor(), QColor(), QColor(),
                          selectionStart, selectionEnd);
     }
 
 }
 
 void QQuickTextNodeEngine::addGlyphsInRange(int rangeStart, int rangeLength,
-                                            const QColor &color, const QColor &backgroundColor,
+                                            const QColor &color, const QColor &backgroundColor, const QColor &decorationColor,
                                             int selectionStart, int selectionEnd)
 {
     QColor oldColor;
@@ -584,6 +604,12 @@ void QQuickTextNodeEngine::addGlyphsInRange(int rangeStart, int rangeLength,
         m_backgroundColor = backgroundColor;
     }
 
+    QColor oldDecorationColor = m_decorationColor;
+    if (decorationColor.isValid()) {
+        oldDecorationColor = m_decorationColor;
+        m_decorationColor = decorationColor;
+    }
+
     bool hasSelection = selectionEnd >= 0
             && selectionStart <= selectionEnd;
 
@@ -627,6 +653,9 @@ void QQuickTextNodeEngine::addGlyphsInRange(int rangeStart, int rangeLength,
         }
     }
 
+    if (decorationColor.isValid())
+        m_decorationColor = oldDecorationColor;
+
     if (backgroundColor.isValid())
         m_backgroundColor = oldBackgroundColor;
 
@@ -638,7 +667,7 @@ void QQuickTextNodeEngine::addBorder(const QRectF &rect, qreal border,
                                      QTextFrameFormat::BorderStyle borderStyle,
                                      const QBrush &borderBrush)
 {
-    QColor color = borderBrush.color();
+    const QColor &color = borderBrush.color();
 
     // Currently we don't support other styles than solid
     Q_UNUSED(borderStyle);
@@ -679,7 +708,8 @@ void QQuickTextNodeEngine::addFrameDecorations(QTextDocument *document, QTextFra
         return;
 
     addBorder(boundingRect.adjusted(frameFormat.leftMargin(), frameFormat.topMargin(),
-                                    -frameFormat.rightMargin(), -frameFormat.bottomMargin()),
+                                    -frameFormat.rightMargin() - borderWidth,
+                                    -frameFormat.bottomMargin() - borderWidth),
               borderWidth, borderStyle, borderBrush);
     if (table != nullptr) {
         int rows = table->rows();
@@ -697,11 +727,9 @@ void QQuickTextNodeEngine::addFrameDecorations(QTextDocument *document, QTextFra
     }
 }
 
-uint qHash(const QQuickTextNodeEngine::BinaryTreeNodeKey &key)
+size_t qHash(const QQuickTextNodeEngine::BinaryTreeNodeKey &key, size_t seed = 0)
 {
-    // Just use the default hash for pairs
-    return qHash(qMakePair(key.fontEngine, qMakePair(key.clipNode,
-                                                     qMakePair(key.color, key.selectionState))));
+    return qHashMulti(seed, key.fontEngine, key.clipNode, key.color, key.selectionState);
 }
 
 void QQuickTextNodeEngine::mergeProcessedNodes(QList<BinaryTreeNode *> *regularNodes,
@@ -800,8 +828,8 @@ void  QQuickTextNodeEngine::addToSceneGraph(QQuickTextNode *parentNode,
     // Then, prepend all selection rectangles to the tree
     for (int i = 0; i < m_selectionRects.size(); ++i) {
         const QRectF &rect = m_selectionRects.at(i);
-
-        parentNode->addRectangleNode(rect, m_selectionColor);
+        if (m_selectionColor.alpha() != 0)
+            parentNode->addRectangleNode(rect, m_selectionColor);
     }
 
     // Add decorations for each node to the tree.
@@ -954,7 +982,17 @@ void QQuickTextNodeEngine::mergeFormats(QTextLayout *textLayout, QVarLengthArray
 
 }
 
-void QQuickTextNodeEngine::addTextBlock(QTextDocument *textDocument, const QTextBlock &block, const QPointF &position, const QColor &textColor, const QColor &anchorColor, int selectionStart, int selectionEnd)
+/*!
+    \internal
+    Adds the \a block from the \a textDocument at \a position if its
+    \l {QAbstractTextDocumentLayout::blockBoundingRect()}{bounding rect}
+    intersects the \a viewport, or if \c viewport is not valid
+    (i.e. use a default-constructed QRectF to skip the viewport check).
+
+    \sa QQuickItem::clipRect()
+ */
+void QQuickTextNodeEngine::addTextBlock(QTextDocument *textDocument, const QTextBlock &block, const QPointF &position,
+                                        const QColor &textColor, const QColor &anchorColor, int selectionStart, int selectionEnd, const QRectF &viewport)
 {
     Q_ASSERT(textDocument);
 #if QT_CONFIG(im)
@@ -969,6 +1007,11 @@ void QQuickTextNodeEngine::addTextBlock(QTextDocument *textDocument, const QText
 
     const QTextCharFormat charFormat = block.charFormat();
     const QRectF blockBoundingRect = textDocument->documentLayout()->blockBoundingRect(block).translated(position);
+    if (viewport.isValid()) {
+        if (!blockBoundingRect.intersects(viewport))
+            return;
+        qCDebug(lcSgText) << "adding block with length" << block.length() << ':' << blockBoundingRect << "in viewport" << viewport;
+    }
 
     if (charFormat.background().style() != Qt::NoBrush)
         m_backgrounds.append(qMakePair(blockBoundingRect, charFormat.background().color()));
@@ -1011,6 +1054,17 @@ void QQuickTextNodeEngine::addTextBlock(QTextDocument *textDocument, const QText
                 break;
             };
 
+            switch (block.blockFormat().marker()) {
+            case QTextBlockFormat::MarkerType::Checked:
+                listItemBullet = QChar(0x2612); // Checked checkbox
+                break;
+            case QTextBlockFormat::MarkerType::Unchecked:
+                listItemBullet = QChar(0x2610); // Unchecked checkbox
+                break;
+            case QTextBlockFormat::MarkerType::NoMarker:
+                break;
+            }
+
             QSizeF size(fontMetrics.horizontalAdvance(listItemBullet), fontMetrics.height());
             qreal xoff = fontMetrics.horizontalAdvance(QLatin1Char(' '));
             if (block.textDirection() == Qt::LeftToRight)
@@ -1088,7 +1142,7 @@ void QQuickTextNodeEngine::addTextBlock(QTextDocument *textDocument, const QText
                 fragmentEnd += preeditLength;
             }
 #endif
-            if (charFormat.background().style() != Qt::NoBrush) {
+            if (charFormat.background().style() != Qt::NoBrush || charFormat.hasProperty(QTextFormat::TextUnderlineColor)) {
                 QTextLayout::FormatRange additionalFormat;
                 additionalFormat.start = textPos - block.position();
                 additionalFormat.length = fragmentEnd - textPos;
@@ -1118,6 +1172,17 @@ void QQuickTextNodeEngine::addTextBlock(QTextDocument *textDocument, const QText
     }
 #endif
 
+    // Add block decorations (so far only horizontal rules)
+    if (block.blockFormat().hasProperty(QTextFormat::BlockTrailingHorizontalRulerWidth)) {
+        auto ruleLength = qvariant_cast<QTextLength>(block.blockFormat().property(QTextFormat::BlockTrailingHorizontalRulerWidth));
+        QRectF ruleRect(0, 0, ruleLength.value(blockBoundingRect.width()), 1);
+        ruleRect.moveCenter(blockBoundingRect.center());
+        const QColor ruleColor = block.blockFormat().hasProperty(QTextFormat::BackgroundBrush)
+                               ? qvariant_cast<QBrush>(block.blockFormat().property(QTextFormat::BackgroundBrush)).color()
+                               : m_textColor;
+        m_lines.append(TextDecoration(QQuickTextNodeEngine::Unselected, ruleRect, ruleColor));
+    }
+
     setCurrentLine(QTextLine()); // Reset current line because the text layout changed
     m_hasContents = true;
 }
